@@ -1,5 +1,6 @@
+from __future__ import annotations
 import abc
-from typing import List, Optional
+from typing import List, Optional, Tuple
 
 import torch
 
@@ -22,6 +23,12 @@ class GeneralFunctionSpecBase(metaclass=abc.ABCMeta):
     def name(self) -> str:
         raise NotImplementedError
 
+    @property
+    def path(self) -> Optional[str]:
+        """The path of this spec relative to RRFS_DIR if this is stored in rrfs.
+        Should be of the form /dir/sub_dir/.../file.py:MyCustomSpec"""
+        return None
+
     def compute_hash_bytes(self) -> bytes:
         """the default implementation should typically be overridden!"""
         return id(self).to_bytes(8, "big", signed=True)
@@ -32,10 +39,11 @@ class GeneralFunctionSpecBase(metaclass=abc.ABCMeta):
         raise NotImplementedError
 
     @abc.abstractmethod
-    def get_shape_info(self, *shapes: List[int]) -> GeneralFunctionShapeInfo:
+    def get_shape_info(self, *shapes: Tuple[int, ...]) -> GeneralFunctionShapeInfo:
         """This should error (exception) if the shapes are invalid and otherwise return GeneralFunctionShapeInfo"""
         raise NotImplementedError
 
-    def serialize(self) -> Optional[str]:
-        """Used in printing. TODO: write spec"""
+    def get_device_dtype_override(self, *device_dtypes: rust_circuit.TorchDeviceDtypeOp) -> Optional[rust_circuit.TorchDeviceDtypeOp]:  # type: ignore
+        """Use this to check the validity of input dtypes/devices and set output device/dtype. If this returns None, the output inherits device/dtype from children.
+        This can error to signal that inputs have invalid dtypes/devices"""
         return None
