@@ -38,11 +38,18 @@ pub trait CircuitNodeComputeInfoImpl: CircuitNode {
     fn device_dtype_extra(&self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + '_> {
         Box::new(std::iter::empty())
     }
+    fn device_dtype_override(&self) -> Option<Result<TorchDeviceDtypeOp>> {
+        None
+    }
     fn symbolic_size_constraints_extra(&self) -> Result<Vec<SymbolicSizeConstraint>> {
         Ok(Vec::new())
     }
 
     fn compute_device_dtype(&self) -> Result<TorchDeviceDtypeOp> {
+        if let Some(z) = self.device_dtype_override() {
+            println!("override device dtype");
+            return z;
+        }
         self.children()
             .map(|c| c.info().device_dtype.clone())
             .chain(self.device_dtype_extra())
