@@ -20,7 +20,7 @@ use crate::{
     circuit_node_private::{CircuitNodeComputeInfoImpl, CircuitNodeHashItems},
     new_rc,
     prelude::*,
-    CachedCircuitInfo, PyCircuitBase, Shape, TensorEvalError,
+    CachedCircuitInfo, CircuitFlags, PyCircuitBase, Shape, TensorEvalError,
 };
 
 macro_rules! circuit_node_auto_leaf_impl {
@@ -227,8 +227,8 @@ impl CircuitNodeComputeInfoImpl for Symbol {
     fn compute_shape(&self) -> Shape {
         self.info.shape.clone() // note: assumes shape has already been initted!
     }
-    fn compute_is_explicitly_computable(&self) -> bool {
-        false
+    fn compute_flags(&self) -> CircuitFlags {
+        CircuitFlags::IS_CONSTANT
     }
 }
 
