@@ -1,4 +1,5 @@
 use std::{
+    cell::RefCell,
     collections::BTreeMap,
     iter::{self, zip},
     ops::Deref,
@@ -9,6 +10,7 @@ use anyhow::{bail, Context, Result};
 use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
+    cached_method,
     compact_data::TinyBoundedBitvec,
     eq_by_big_hash::EqByBigHash,
     impl_eq_by_big_hash, pycall, python_error_exception,
@@ -23,8 +25,8 @@ use smallvec::SmallVec;
 use thiserror::Error;
 
 use crate::{
-    deep_map_fallible_pre_new_children, prelude::*, Add, Concat, Cumulant, Einsum, GeneralFunction,
-    Index, Module, ModuleArgSpec, ModuleSpec, Rearrange, Scatter, SetSymbolicShape, Symbol, Tag,
+    prelude::*, Add, Concat, Cumulant, Einsum, GeneralFunction, Index, Module, ModuleArgSpec,
+    ModuleSpec, Rearrange, Scatter, SetSymbolicShape, Symbol, Tag,
 };
 
 #[pyfunction]
@@ -37,7 +39,7 @@ pub fn expand_node_py(circuit: CircuitRc, inputs: Vec<CircuitRc>) -> Result<Circ
 
 #[derive(Clone, Debug)]
 pub struct ReplaceMapRc {
-    replace_map: Arc<HashMap<CircuitRc, CircuitRc>>,
+    replace_map: Arc<BTreeMap<CircuitRc, CircuitRc>>,
     hash: HashBytes,
 }
 
@@ -48,7 +50,7 @@ impl Default for ReplaceMapRc {
 }
 
 impl Deref for ReplaceMapRc {
-    type Target = HashMap<CircuitRc, CircuitRc>;
+    type Target = BTreeMap<CircuitRc, CircuitRc>;
 
     fn deref(&self) -> &Self::Target {
         &self.replace_map
@@ -56,7 +58,7 @@ impl Deref for ReplaceMapRc {
 }
 
 impl ReplaceMapRc {
-    pub fn new(replace_map: HashMap<CircuitRc, CircuitRc>) -> Self {
+    pub fn new(replace_map: BTreeMap<CircuitRc, CircuitRc>) -> Self {
         let mut hasher = blake3::Hasher::new();
         for (k, v) in replace_map.iter() {
             hasher.update(&k.hash());
@@ -71,6 +73,36 @@ impl ReplaceMapRc {
     pub fn into_map(self) -> <Self as Deref>::Target {
         arc_unwrap_or_clone(self.replace_map)
     }
+
+    pub fn extend_into(&self, new: impl IntoIterator<Item = (CircuitRc, CircuitRc)>) -> Self {
+        Self::new(
+            self.iter()
+                .map(|(a, b)| (a.clone(), b.clone()))
+                .chain(new)
+                .collect(),
+        )
+    }
+
+    /// handle the module case
+    pub fn per_child(&self, circuit: &CircuitRc) -> Vec<Self> {
+        if let Some(m) = circuit.as_module() {
+            if m.spec
+                .arg_specs
+                .iter()
+                .any(|arg_spec| self.contains_key(&arg_spec.symbol.crc()))
+            {
+                let mut removed_map = (*self.replace_map).clone();
+                for arg_spec in &m.spec.arg_specs {
+                    removed_map.remove(&arg_spec.symbol.crc());
+                }
+
+                return iter::once(Self::new(removed_map))
+                    .chain(iter::repeat(self).take(m.nodes.len()).cloned())
+                    .collect();
+            }
+        }
+        vec![self.clone(); circuit.num_children()]
+    }
 }
 
 impl EqByBigHash for ReplaceMapRc {
@@ -801,42 +833,115 @@ pub fn replace_expand_bottom_up_py(circuit: CircuitRc, f: PyObject) -> Result<Ci
     replace_expand_bottom_up(circuit, |x| pycall!(f, (x.clone(),)))
 }
 
-pub fn replace_expand_bottom_up<F>(circuit: CircuitRc, replacer: F) -> Result<CircuitRc>
-where
-    F: Fn(CircuitRc) -> Option<CircuitRc>,
-{
-    replace_expand_bottom_up_impl(circuit, &replacer, Default::default())
+pub trait ReplaceForExpand {
+    fn replace(&mut self, circuit: CircuitRc) -> Option<CircuitRc>;
+    fn done(&mut self, _circuit: CircuitRc) -> bool {
+        false
+    }
 }
 
-pub fn replace_expand_bottom_up_impl<F>(
-    circuit: CircuitRc,
-    replacer: &F,
-    extra_replacements: BTreeMap<CircuitRc, CircuitRc>,
-) -> Result<CircuitRc>
-where
-    F: Fn(CircuitRc) -> Option<CircuitRc>,
-{
-    let recursor = |circuit: CircuitRc, new_children: &Vec<CircuitRc>| -> Result<CircuitRc> {
-        if let Some(replaced) = replacer(circuit.clone()) {
+pub struct ReplaceForExpandFn<F: FnMut(CircuitRc) -> Option<CircuitRc>>(pub F);
+
+impl<F: FnMut(CircuitRc) -> Option<CircuitRc>> ReplaceForExpand for ReplaceForExpandFn<F> {
+    fn replace(&mut self, circuit: CircuitRc) -> Option<CircuitRc> {
+        self.0(circuit)
+    }
+}
+
+pub struct ReplaceExpander<F: ReplaceForExpand> {
+    replacer: F,
+    // TODO: this holds onto array tensors forever!
+    // We could try to use 'weak' for the cache, but this is a bit sad... (drops a bunch of stuff instantly)
+    // Fix me!
+    cache: cached::UnboundCache<(HashBytes, HashBytes), CircuitRc>,
+}
+
+impl<F: FnMut(CircuitRc) -> Option<CircuitRc>> ReplaceExpander<ReplaceForExpandFn<F>> {
+    fn new_fn(replacer: F) -> Self {
+        Self::new(ReplaceForExpandFn(replacer))
+    }
+}
+
+pub struct ReplaceForExpandNoop;
+
+impl ReplaceForExpand for ReplaceForExpandNoop {
+    fn replace(&mut self, _circuit: CircuitRc) -> Option<CircuitRc> {
+        None
+    }
+    fn done(&mut self, _circuit: CircuitRc) -> bool {
+        true
+    }
+}
+
+type MapReplaceExpander = ReplaceExpander<ReplaceForExpandNoop>;
+impl MapReplaceExpander {
+    pub fn new_noop() -> Self {
+        Self {
+            replacer: ReplaceForExpandNoop,
+            cache: cached::UnboundCache::new(),
+        }
+    }
+}
+
+impl<F: ReplaceForExpand> ReplaceExpander<F> {
+    pub fn new(replacer: F) -> Self {
+        Self {
+            replacer,
+            cache: cached::UnboundCache::new(),
+        }
+    }
+
+    pub fn replace_expand(&mut self, circuit: CircuitRc) -> Result<CircuitRc> {
+        self.replace_expand_with_map(circuit, &Default::default())
+    }
+
+    #[apply(cached_method)]
+    #[self_id(self_)]
+    #[key((circuit.info().hash, extra_replacements.hash()))]
+    #[use_try]
+    #[cache_expr(cache)]
+    pub fn replace_expand_with_map(
+        &mut self,
+        circuit: CircuitRc,
+        extra_replacements: &ReplaceMapRc,
+    ) -> Result<CircuitRc> {
+        if extra_replacements.is_empty() && self_.replacer.done(circuit.clone()) {
+            return Ok(circuit);
+        }
+        if let Some(replaced) = self_.replacer.replace(circuit.clone()) {
             return Ok(replaced);
         }
         if let Some(replaced) = extra_replacements.get(&circuit) {
             return Ok(replaced.clone());
         }
-        expand_node(circuit, new_children, |c, rep, _child_idx| {
-            // This recursion is a bit gross + uncached (caching requires caching by extra_replacements ofc, so a bit more complex) : /
-            replace_expand_bottom_up_impl(
-                c,
-                replacer,
-                extra_replacements
-                    .iter()
-                    .map(|(a, b)| (a.clone(), b.clone()))
-                    .chain(rep)
-                    .collect(),
-            )
+
+        // TODO: optimize this as needed
+        let new_children = circuit
+            .children()
+            .zip(extra_replacements.per_child(&circuit))
+            .map(|(c, rep)| self_.replace_expand_with_map(c, &rep))
+            .collect::<Result<_>>()?;
+
+        expand_node(circuit, &new_children, |c, rep, _child_idx| {
+            self_.replace_expand_with_map(c, &extra_replacements.extend_into(rep))
         })
-    };
-    deep_map_fallible_pre_new_children(circuit, recursor)
+    }
+}
+
+pub fn replace_expand_bottom_up<F>(circuit: CircuitRc, replacer: F) -> Result<CircuitRc>
+where
+    F: Fn(CircuitRc) -> Option<CircuitRc>,
+{
+    ReplaceExpander::new_fn(replacer).replace_expand(circuit)
+}
+
+thread_local! {
+    static MAP_REPLACE_EXPANDER: RefCell<MapReplaceExpander> =
+        RefCell::new(MapReplaceExpander::new_noop());
+}
+
+pub fn map_replace_expand(circuit: CircuitRc, map: &ReplaceMapRc) -> Result<CircuitRc> {
+    MAP_REPLACE_EXPANDER.with(|f| f.borrow_mut().replace_expand_with_map(circuit, map))
 }
 
 #[test]
