@@ -1,12 +1,12 @@
 use std::{fmt, iter::zip};
 
-use anyhow::{bail, Context, Result};
+use anyhow::{anyhow, bail, Context, Result};
 use macro_rules_attribute::apply;
 use once_cell::sync::Lazy;
 use pyo3::{
     exceptions::PyValueError,
     once_cell::{GILLazy, GILLazyPy},
-    prelude::*,
+    prelude::{PyAny, *},
     types::PyTuple,
 };
 use rand::Rng;
@@ -14,8 +14,10 @@ use regex::Regex;
 use rr_util::{
     compact_data::TinyBoundedBitvec,
     py_types::{assert_tensors_close, ExtraPySelfOps, PyShape, Tensor, PY_UTILS},
-    python_error_exception, simple_from,
-    tensor_util::{check_canon_idxs, Shape, TorchDeviceDtype},
+    pycall, python_error_exception, simple_from,
+    tensor_util::{
+        check_canon_idxs, MiscInputError, Shape, TorchDeviceDtype, TorchDeviceDtypeOp, TorchDtype,
+    },
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use thiserror::Error;
@@ -114,6 +116,12 @@ pub trait SpecTrait: fmt::Debug + ToPyObject {
         false
     }
     fn serialize(&self) -> Result<Option<String>>;
+    fn get_device_dtype_override(
+        &self,
+        device_dtypes: &[TorchDeviceDtypeOp],
+    ) -> Option<Result<TorchDeviceDtypeOp>> {
+        None
+    } // returning None means use normal device_dtype inheritance. if Some, then error if dtypes are incorrect and dtype if correct
 }
 
 #[pyclass]
@@ -376,7 +384,18 @@ impl SpecTrait for GeneralFunctionIndexSpec {
     //     Ok(self.get_jacobians.is_some())
     // }
     // fn get_jacobians(&self, func: &GeneralFunction) -> Result<Option<Vec<Circuit>>> {}
-
+    fn get_device_dtype_override(
+        &self,
+        device_dtypes: &[TorchDeviceDtypeOp],
+    ) -> Option<Result<TorchDeviceDtypeOp>> {
+        if let Some(dev1) = device_dtypes[0].device && let Some(dev2) = device_dtypes[1].device && dev1!=dev2{
+            return Some(Err(MiscInputError::ChildrenMultipleDevices { a: device_dtypes[0].device, b: device_dtypes[1].device}.into()))
+        }
+        if let Some(dtype) = device_dtypes[1].dtype && dtype != TorchDtype::int64 {
+            return Some(Err(MiscInputError::IndexDtypeNotI64 {}.into()));
+        }
+        Some(Ok(device_dtypes[0]))
+    }
     fn name(&self) -> String {
         format!(
             "gen_index_at_{}{}{}",
@@ -392,11 +411,95 @@ impl SpecTrait for GeneralFunctionIndexSpec {
         Ok(Some(self.name()))
     }
 }
+
+#[pyclass]
+#[derive(Debug, Clone)]
+pub struct GeneralFunctionSetDDSpec {
+    #[pyo3(get)]
+    input_required_compatibility: TorchDeviceDtypeOp,
+    #[pyo3(get)]
+    output: TorchDeviceDtypeOp,
+}
+
+simple_from!(|x: GeneralFunctionSetDDSpec| -> GeneralFunctionSpec {
+    GeneralFunctionSpec::SetDD(x)
+});
+
+impl ToPyObject for GeneralFunctionSetDDSpec {
+    fn to_object(&self, py: Python<'_>) -> PyObject {
+        self.clone().into_py(py)
+    }
+}
+
+impl SpecTrait for GeneralFunctionSetDDSpec {
+    fn compute_hash(&self) -> HashBytes {
+        let mut hasher = blake3::Hasher::new();
+        self.input_required_compatibility.hash(&mut hasher);
+        self.output.hash(&mut hasher);
+        *hasher.finalize().as_bytes()
+    }
+
+    fn function(&self, tensors: &[Tensor], _device_dtype: &TorchDeviceDtype) -> Result<Tensor> {
+        pycall!(
+            PY_UTILS.cast_tensor,
+            (tensors[0].clone(), self.output),
+            anyhow
+        )
+    }
+
+    fn get_shape_info(&self, shapes: &[Shape]) -> Result<GeneralFunctionShapeInfo> {
+        if shapes.len() != 1 {
+            bail!(GeneralFunctionShapeError::WrongNumShapes {
+                got: shapes.len(),
+                expected: 1
+            });
+        }
+        Ok(GeneralFunctionShapeInfo {
+            shape: shapes[0].clone(),
+            num_non_batchable_output_dims: 0,
+            input_batchability: vec![false].try_into().unwrap(),
+        })
+    }
+    fn get_device_dtype_override(
+        &self,
+        device_dtypes: &[TorchDeviceDtypeOp],
+    ) -> Option<Result<TorchDeviceDtypeOp>> {
+        Some(
+            self.input_required_compatibility
+                .combine(device_dtypes[0])
+                .map_err(|_e| {
+                    MiscInputError::CastIncompatibleDeviceDtype {
+                        required: self.input_required_compatibility,
+                        actual: device_dtypes[0],
+                    }
+                    .into()
+                })
+                .map(|_| self.output.override_other(device_dtypes[0])),
+        )
+    }
+
+    // fn has_jacobian(&self) -> Result<bool> {
+    //     Ok(self.get_jacobians.is_some())
+    // }
+    // fn get_jacobians(&self, func: &GeneralFunction) -> Result<Option<Vec<Circuit>>> {}
+
+    fn name(&self) -> String {
+        format!(
+            "cast_{:?}_{:?}",
+            self.input_required_compatibility, self.output
+        )
+    }
+    fn serialize(&self) -> Result<Option<String>> {
+        Ok(Some(self.name()))
+    }
+}
+
 #[derive(Clone)]
 pub struct PyWrap {
     ob: PyObject,
     hash: HashBytes,
     name: String,
+    path: Option<String>,
 }
 
 impl fmt::Debug for PyWrap {
@@ -404,6 +507,7 @@ impl fmt::Debug for PyWrap {
         f.debug_struct("PyWrap")
             .field("ob", &self.ob)
             .field("name", &self.name)
+            .field("path", &self.path)
             .finish_non_exhaustive()
     }
 }
@@ -444,8 +548,23 @@ impl<'source> pyo3::FromPyObject<'source> for Box<PyWrap> {
         })?;
         hasher.update(&bytes);
 
+        let name: String =
+            Python::with_gil(|py| -> Result<_> { Ok(ob.getattr(py, "name")?.extract(py)?) })?;
+
+        if name.contains(" at ") {
+            return Err(anyhow!("Function names can't contain the substring ` at ` because it could interfere with parsing.").into());
+        }
+
+        let path: Option<String> =
+            Python::with_gil(|py| -> Result<_> { Ok(ob.getattr(py, "path")?.extract(py)?) })?;
+
+        if let Some(p) = &path && (!p.starts_with("/") || !p.contains(".py:")) {
+            return Err(anyhow!("Function path of the form /dir/sub_dir/.../file.py:MyCustomSpec. (relative to rrfs)").into());
+        }
+
         Ok(Box::new(PyWrap {
-            name: Python::with_gil(|py| -> Result<_> { Ok(ob.getattr(py, "name")?.extract(py)?) })?,
+            name,
+            path,
             ob,
             hash: *hasher.finalize().as_bytes(),
         }))
@@ -488,11 +607,43 @@ impl SpecTrait for Box<PyWrap> {
                 .extract(py)?)
         })
     }
+    fn get_device_dtype_override(
+        &self,
+        device_dtypes: &[TorchDeviceDtypeOp],
+    ) -> Option<Result<TorchDeviceDtypeOp>> {
+        let resulty: Result<Option<TorchDeviceDtypeOp>> = Python::with_gil(|py| {
+            Ok(self
+                .ob
+                .call_method1(
+                    py,
+                    "get_device_dtype_override",
+                    // PyShape here is important!
+                    PyTuple::new(
+                        py,
+                        device_dtypes
+                            .iter()
+                            .map(|x| x.into_py(py))
+                            .collect::<Vec<_>>(),
+                    ),
+                )?
+                .extract(py)?)
+        });
+        if let Err(e) = resulty {
+            return Some(Err(e));
+        }
+        return resulty.unwrap().map(|x| Ok(x));
+    }
     fn name(&self) -> String {
         self.name.clone()
     }
     fn serialize(&self) -> Result<Option<String>> {
-        Python::with_gil(|py| Ok(self.ob.call_method0(py, "serialize")?.extract(py)?))
+        match &self.path {
+            Some(p) => Ok(Some(format!("{} at {}", &self.name, p))),
+            None => bail!(anyhow!(format!(
+                "Can't serialize function {}: no path given",
+                &self.name
+            ))),
+        }
     }
 }
 
@@ -500,6 +651,7 @@ impl SpecTrait for Box<PyWrap> {
 pub enum GeneralFunctionSpec {
     Simple(GeneralFunctionSimpleSpec),
     Index(GeneralFunctionIndexSpec),
+    SetDD(GeneralFunctionSetDDSpec),
     Py(Box<PyWrap>), // boxed bc it was long pole for Circuit size
 }
 
@@ -509,6 +661,7 @@ impl GeneralFunctionSpec {
         match self {
             Self::Simple(x) => x,
             Self::Index(x) => x,
+            Self::SetDD(x) => x,
             Self::Py(x) => x,
         }
     }
@@ -535,6 +688,12 @@ impl SpecTrait for GeneralFunctionSpec {
     fn get_shape_info(&self, shapes: &[Shape]) -> Result<GeneralFunctionShapeInfo> {
         self.as_trait_obj().get_shape_info(shapes)
     }
+    fn get_device_dtype_override(
+        &self,
+        device_dtypes: &[TorchDeviceDtypeOp],
+    ) -> Option<Result<TorchDeviceDtypeOp>> {
+        self.as_trait_obj().get_device_dtype_override(device_dtypes)
+    }
     // fn has_jacobian(&self) -> Result<bool>;
     // fn get_jacobians(&self, func: &GeneralFunction) -> Result<Option<Vec<Circuit>>>;
     fn name(&self) -> String {
@@ -613,6 +772,16 @@ impl CircuitNodeComputeInfoImpl for GeneralFunction {
     fn compute_shape(&self) -> Shape {
         self.info.shape.clone() // set in try new!
     }
+
+    fn device_dtype_override(&self) -> Option<Result<TorchDeviceDtypeOp>> {
+        self.spec.get_device_dtype_override(
+            &self
+                .nodes
+                .iter()
+                .map(|x| x.info().device_dtype)
+                .collect::<Vec<_>>(),
+        )
+    }
 }
 
 impl CircuitNodeHashItems for GeneralFunction {
@@ -712,8 +881,19 @@ impl GeneralFunction {
         parse_string: String,
         name: Option<String>,
     ) -> Result<Self> {
-        // TODO: allow py wrap to do something here!
-        Self::new_by_name(nodes, parse_string, name)
+        Python::with_gil(|_py| {
+            println!("{parse_string}");
+        });
+
+        if parse_string.contains(" at ") {
+            let mut split = parse_string.splitn(2, " at ");
+            let name = convert_op_str_to_format_err(split.next(), &parse_string)?;
+            let path = convert_op_str_to_format_err(split.next(), &parse_string)?;
+
+            Self::new_by_path(nodes, path, Some(name))
+        } else {
+            Self::new_by_name(nodes, parse_string, name)
+        }
     }
 
     #[staticmethod]
@@ -761,6 +941,38 @@ impl GeneralFunction {
         GeneralFunction::try_new(nodes, spec, name).map(Some)
     }
 
+    #[staticmethod]
+    #[args(nodes = "*", path, name = "None")]
+    pub fn new_by_path(nodes: Vec<CircuitRc>, path: String, name: Option<String>) -> Result<Self> {
+        let mut split = path.splitn(2, ":");
+        let file_path = convert_op_str_to_format_err(split.next(), &path)?;
+        let spec_name = convert_op_str_to_format_err(split.next(), &path)?;
+
+        let spec: GeneralFunctionSpec =
+            Python::with_gil(|py| -> Result<GeneralFunctionSpec, pyo3::PyErr> {
+                let rrfs_dir = py
+                    .import("interp.tools.rrfs")?
+                    .getattr("RRFS_DIR")?
+                    .extract::<String>()?;
+
+                // adapted from https://stackoverflow.com/questions/67631/how-can-i-import-a-module-dynamically-given-the-full-path
+
+                let importlib_util = py.import("importlib.util")?;
+                let spec = importlib_util.call_method1(
+                    "spec_from_file_location",
+                    ("user_funcs", rrfs_dir + &file_path),
+                )?;
+                let function_spec_module =
+                    importlib_util.call_method1("module_from_spec", (spec,))?;
+                spec.getattr("loader")?
+                    .call_method1("exec_module", (function_spec_module,))?;
+                let spec_cls = function_spec_module.getattr(&*spec_name)?;
+                spec_cls.call0()?.extract()
+            })?;
+
+        GeneralFunction::try_new(nodes, spec, name)
+    }
+
     #[staticmethod]
     #[args(batch_x = "false", check_index_ints = "true", name = "None")]
     pub fn gen_index(
@@ -779,6 +991,25 @@ impl GeneralFunction {
         .into();
         Self::try_new(vec![x, index], spec, name)
     }
+    #[staticmethod]
+    #[args(
+        input_required_compatibility = "Default::default()",
+        output = "Default::default()",
+        name = "None"
+    )]
+    pub fn new_cast(
+        x: CircuitRc,
+        input_required_compatibility: TorchDeviceDtypeOp,
+        output: TorchDeviceDtypeOp,
+        name: Option<String>,
+    ) -> Result<Self> {
+        let spec = GeneralFunctionSetDDSpec {
+            input_required_compatibility,
+            output,
+        }
+        .into();
+        Self::try_new(vec![x], spec, name)
+    }
 }
 
 /// I now think this maybe should have been written in python.
@@ -1292,6 +1523,17 @@ fn test_index_general_function() -> Result<()> {
     Ok(())
 }
 
+fn convert_op_str_to_format_err(op_s: Option<&str>, path: &String) -> Result<String> {
+    if let Some(s) = op_s {
+        Ok(s.to_string())
+    } else {
+        Err(anyhow!(format!(
+            "failed to parse general function from string: {}",
+            path
+        )))
+    }
+}
+
 #[apply(python_error_exception)]
 #[base_error_name(GeneralFunctionShape)]
 #[base_exception(PyValueError)]
