@@ -39,7 +39,10 @@ use rr_util::{
     rearrange_spec::{OpSize, PyCountsAtAxes, RInnerInts, RearrangeSpec},
     sv,
     symbolic_size::{SymbolicSizeConstraints, SymbolicSizeProduct},
-    tensor_util::{check_canon_idxs, Shape, TensorIndex, TorchDeviceDtype, TorchDeviceDtypeOp},
+    tensor_util::{
+        check_canon_idxs, Shape, TensorIndex, TorchDevice, TorchDeviceDtype, TorchDeviceDtypeOp,
+        TorchDtype,
+    },
     tu8v,
     util::{
         arc_ref_clone, arc_unwrap_or_clone, AsOp, EinsumAxes, HashBytes, HashBytesToPy, NamedAxes,
@@ -2032,6 +2035,11 @@ impl PyCircuitBase {
         self.0.intermediate_cost_bound()
     }
 
+    #[getter]
+    fn symbolic_size_constraints(&self) -> SymbolicSizeConstraints {
+        self.info().symbolic_size_constraints.clone()
+    }
+
     fn children(&self) -> Vec<CircuitRc> {
         self.0.children().collect()
     }
@@ -2128,10 +2136,18 @@ impl PyCircuitBase {
     fn total_arrayconstant_size(&self) -> BigUint {
         total_arrayconstant_size(self.0.clone())
     }
-
-    fn get_device_dtype(&self) -> TorchDeviceDtypeOp {
+    #[getter]
+    fn device_dtype(&self) -> TorchDeviceDtypeOp {
         self.0.info().device_dtype.clone()
     }
+    #[getter]
+    fn device(&self) -> Option<TorchDevice> {
+        self.0.info().device_dtype.device
+    }
+    #[getter]
+    fn dtype(&self) -> Option<TorchDtype> {
+        self.0.info().device_dtype.dtype
+    }
 
     fn get_compatible_device_dtype(&self) -> TorchDeviceDtype {
         get_compatible_dtype(&***self)
