@@ -1,7 +1,7 @@
 use std::{
     hash::{Hash, Hasher},
     iter::{once, zip},
-    sync::{Arc, Weak},
+    // sync::{Arc, Weak},
 };
 
 use anyhow::{bail, Context, Result};
@@ -14,7 +14,7 @@ use rr_util::{
     rearrange_spec::RearrangeSpec,
     symbolic_size::SymbolicSizeProduct,
     tensor_util::{right_align_shapes, Shape, TorchDeviceDtype},
-    util::{counts_g_1, is_unique, unique_to_appearance},
+    util::{counts_g_1, is_unique},
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet, FxHasher};
 use thiserror::Error;
@@ -25,12 +25,12 @@ use crate::{
     circuit_node_private::{CircuitNodeHashItems, CircuitNodePrivate, CircuitNodeSetNonHashInfo},
     circuit_utils::OperatorPriority,
     deep_map, deep_map_op_context_preorder_stoppable, deep_map_preorder_unwrap, deep_map_unwrap,
-    expand_node::{replace_expand_bottom_up, ExpandError},
+    expand_node::{ExpandError, ReplaceExpander, ReplaceMapRc},
     named_axes::{deep_strip_axis_names, propagate_named_axes},
     new_rc_unwrap,
     prelude::*,
-    visit_circuit, visit_circuit_unwrap, CachedCircuitInfo, HashBytes, PyCircuitBase, Rearrange,
-    Symbol, TensorEvalError,
+    visit_circuit_unwrap, CachedCircuitInfo, HashBytes, PyCircuitBase, Rearrange, Symbol,
+    TensorEvalError,
 };
 
 /// can also be thought of as lambda from lambda calculus (expression with free variables + list of these variables)
@@ -45,6 +45,27 @@ pub struct ModuleSpec {
 
 type ModuleSpecHashable = (HashBytes, Vec<ModuleArgSpecHashable>);
 
+pub fn are_args_used<'a, I: IntoIterator<Item = &'a ModuleArgSpec>>(
+    circuit: &CircuitRc,
+    arg_specs: I,
+) -> Vec<bool>
+where
+    <I as IntoIterator>::IntoIter: DoubleEndedIterator,
+{
+    // right most binding has precedence, prior bindings unused
+    let mut out: Vec<_> = arg_specs
+        .into_iter()
+        .rev()
+        .scan(HashSet::default(), |prior_syms, spec| {
+            let out = (!prior_syms.contains(&spec.symbol)) && has_free_sym(&circuit, &spec.symbol);
+            prior_syms.insert(&spec.symbol);
+            Some(out)
+        })
+        .collect();
+    out.reverse();
+    out
+}
+
 impl ModuleSpec {
     pub const EXPAND_PLACEHOLDER_UUID: Uuid = uuid!("741ba404-eec3-4ac9-b6ce-062e903fb033");
 
@@ -134,12 +155,17 @@ impl ModuleSpec {
             .aligned_batch_shape(nodes)
             .context("batch shapes didn't match for module")?;
 
-        let out = replace_expand_bottom_up(self.circuit.clone(), |c| {
-            self.arg_specs
-                .iter()
-                .position(|x| x.symbol.info().hash == c.info().hash)
-                .map(|i| nodes[i].clone())
-        })?;
+        // TODO: fix this being uncached!
+        let out = ReplaceExpander::new_noop().replace_expand_with_map(
+            self.circuit.clone(),
+            &ReplaceMapRc::new(
+                self.arg_specs
+                    .iter()
+                    .zip(nodes)
+                    .map(|(arg_spec, node)| (arg_spec.symbol.crc(), node.clone()))
+                    .collect(),
+            ),
+        )?;
 
         assert!(out.ndim() >= self.circuit.ndim());
         let out_batch_rank = out.ndim() - self.circuit.ndim();
@@ -164,12 +190,20 @@ impl ModuleSpec {
     }
 
     pub fn expand_shape(&self, shapes: &Vec<Shape>) -> Result<(CircuitRc, Vec<HashBytes>)> {
+        let key = (self.get_hashable(), shapes.clone());
         if let Some(result) = MODULE_EXPANSIONS_SHAPE.with(|cache| {
-            cache
-                .borrow()
-                .get(&(self.get_hashable(), shapes.clone()))
-                .map(|(w, shapes)| w.upgrade().map(|w| (CircuitRc(w), shapes.clone())))
-                .flatten()
+            let borrowed = cache.borrow();
+            if let Some((w, shapes)) = borrowed.get(&key) {
+                return Some((w.clone(), shapes.clone()));
+
+                // if let Some(w) = w.upgrade() {
+                //     return Some((CircuitRc(w), shapes.clone()));
+                // } else {
+                //     drop(borrowed);
+                //     cache.borrow_mut().remove(&key);
+                // }
+            }
+            None
         }) {
             return Ok(result);
         }
@@ -189,7 +223,11 @@ impl ModuleSpec {
         MODULE_EXPANSIONS_SHAPE.with(|cache| {
             cache.borrow_mut().insert(
                 (self.get_hashable(), shapes.clone()),
-                (Arc::downgrade(&result.0), out_bytes.clone()),
+                (
+                    // Arc::downgrade(&result.0)
+                    result.clone(),
+                    out_bytes.clone(),
+                ),
             )
         });
         Ok((result, out_bytes))
@@ -208,41 +246,23 @@ impl ModuleSpec {
         );
 
         if let Some(result) = MODULE_EXPANSIONS.with(|cache| {
-            cache
-                .borrow()
-                .get(&key)
-                .map(|x| x.upgrade().map(CircuitRc))
-                .flatten()
+            let borrowed = cache.borrow();
+            if let Some(w) = borrowed.get(&key) {
+                return Some(w.clone());
+                // if let Some(w) = w.upgrade() {
+                //     return Some(CircuitRc(w));
+                // } else {
+                //     drop(borrowed);
+                //     cache.borrow_mut().remove(&key);
+                // }
+            }
+            None
         }) {
             return Ok(result);
         }
 
         for (arg_spec, sub_for_sym) in self.arg_specs.iter().zip(nodes) {
-            check_valid_sub(self.circuit.clone(), &arg_spec.symbol, sub_for_sym.clone())?
-        }
-        for idxs in unique_to_appearance(&self.arg_specs.iter().map(|x| x.symbol.ident()).collect())
-            .into_values()
-        {
-            if idxs.len() <= 1 {
-                continue;
-            }
-            if idxs
-                .iter()
-                .map(|i| &nodes[*i])
-                .collect::<HashSet<_>>()
-                .len()
-                > 1
-            {
-                bail!(SubstitutionError::BindsEquivalentSymbolsToDifferentNodes {
-                    symbols: self
-                        .arg_specs
-                        .iter()
-                        .map(|arg_spec| arg_spec.symbol.clone())
-                        .collect(),
-                    nodes: nodes.clone(),
-                    idxs
-                });
-            }
+            check_valid_sub(&self.circuit, &arg_spec.symbol, sub_for_sym)?
         }
 
         let shapes = nodes.iter().map(|x| x.info().shape.clone()).collect();
@@ -282,7 +302,12 @@ impl ModuleSpec {
                 self, nodes
             )
         })?;
-        MODULE_EXPANSIONS.with(|cache| cache.borrow_mut().insert(key, Arc::downgrade(&result.0)));
+        MODULE_EXPANSIONS.with(|cache| {
+            cache.borrow_mut().insert(
+                key,
+                result.clone(), // Arc::downgrade(&result.0)
+            )
+        });
         Ok(result)
     }
 
@@ -333,16 +358,16 @@ pub fn all_free(circuit: CircuitRc) -> Vec<Symbol> {
 #[pymethods]
 impl ModuleSpec {
     #[new]
-    #[args(check_all_args_present = "true", check_unique_arg_names = "true")]
+    #[args(check_all_inputs_used = "true", check_unique_arg_names = "true")]
     pub fn new(
         circuit: CircuitRc,
         arg_specs: Vec<ModuleArgSpec>,
-        check_all_args_present: bool,
+        check_all_inputs_used: bool,
         check_unique_arg_names: bool,
     ) -> Result<Self> {
         let out = Self { circuit, arg_specs };
-        if check_all_args_present {
-            out.check_all_args_present()?;
+        if check_all_inputs_used {
+            out.check_all_inputs_used()?;
         }
         if check_unique_arg_names {
             out.check_unique_arg_names()?;
@@ -350,11 +375,12 @@ impl ModuleSpec {
         Ok(out)
     }
 
-    pub fn check_all_args_present(&self) -> Result<()> {
+    pub fn check_all_inputs_used(&self) -> Result<()> {
         let missing_symbols: HashSet<_> = self
-            .arg_specs
-            .iter()
-            .filter_map(|x| (!has_sym(&self.circuit, &x.symbol)).then(|| x.symbol.clone()))
+            .are_args_used()
+            .into_iter()
+            .zip(&self.arg_specs)
+            .filter_map(|(used, arg_spec)| (!used).then(|| arg_spec.symbol.clone()))
             .collect();
         if !missing_symbols.is_empty() {
             bail!(ConstructError::ModuleSomeArgsNotPresent {
@@ -413,11 +439,11 @@ impl ModuleSpec {
     }
 
     #[staticmethod]
-    #[args(check_all_args_present = "true", check_unique_arg_names = "true")]
+    #[args(check_all_inputs_used = "true", check_unique_arg_names = "true")]
     pub fn new_extract(
         circuit: CircuitRc,
         arg_specs: Vec<(CircuitRc, ModuleArgSpec)>,
-        check_all_args_present: bool,
+        check_all_inputs_used: bool,
         check_unique_arg_names: bool,
     ) -> Result<Self> {
         let mut new_arg_specs: Vec<Option<ModuleArgSpec>> = vec![None; arg_specs.len()];
@@ -448,7 +474,7 @@ impl ModuleSpec {
             &mut Default::default(),
         )
         .unwrap_or(circuit);
-        let new_arg_specs: Vec<ModuleArgSpec> = if check_all_args_present {
+        let new_arg_specs: Vec<ModuleArgSpec> = if check_all_inputs_used {
             let er = new_arg_specs
                 .iter()
                 .cloned()
@@ -468,7 +494,7 @@ impl ModuleSpec {
                 .unwrap()
         };
         // maybe we should use no_check_args for speed
-        let out = Self::new(spec_circuit, new_arg_specs, check_all_args_present, false)
+        let out = Self::new(spec_circuit, new_arg_specs, check_all_inputs_used, false)
             .expect("method should guarantee valid");
         if check_unique_arg_names {
             // check after instead of in new so we can .expect on other errors in new
@@ -496,6 +522,10 @@ impl ModuleSpec {
         Ok(Self { circuit, arg_specs })
     }
 
+    pub fn are_args_used(&self) -> Vec<bool> {
+        are_args_used(&self.circuit, &self.arg_specs)
+    }
+
     pub fn __repr__(&self) -> String {
         format!("{:?}", self)
     }
@@ -1015,133 +1045,179 @@ pub fn any_children_with_symbolic_sizes(circuit: CircuitRc) -> bool {
 // adhoc anyfound
 // TODO: do this batchedly over symbols while still caching individually?
 #[derive(Clone, Debug)]
-pub struct HasSym {
-    cache: cached::UnboundCache<(HashBytes, HashBytes, bool), bool>,
+pub struct HasFreeSym {
+    cache: cached::UnboundCache<(HashBytes, HashBytes), bool>,
 }
-impl Default for HasSym {
+impl Default for HasFreeSym {
     fn default() -> Self {
         Self {
             cache: cached::UnboundCache::new(),
         }
     }
 }
-impl HasSym {
+impl HasFreeSym {
     #[apply(cached_method)]
     #[self_id(self_)]
-    #[key((circuit.info().hash, sym.info().hash, by_ident))]
+    #[key((circuit.info().hash, sym.info().hash))]
     #[cache_expr(cache)]
-    pub fn has_sym(&mut self, circuit: &CircuitRc, sym: &Symbol, by_ident: bool) -> bool {
-        if (by_ident && circuit.as_symbol().map(|x| x.ident()) == Some(sym.ident()))
-            || (!by_ident && circuit.as_symbol() == Some(sym))
-        {
-            true
-        } else {
-            circuit.children().any(|c| self_.has_sym(&c, sym, by_ident))
+    pub fn has_free_sym(&mut self, circuit: &CircuitRc, sym: &Symbol) -> bool {
+        if circuit.as_symbol() == Some(sym) {
+            return true;
         }
+
+        if let Some(m) = circuit.as_module() {
+            if m.spec
+                .arg_specs
+                .iter()
+                .any(|arg_spec| &arg_spec.symbol == sym)
+            {
+                // skip spec circuit if bound
+                return m.nodes.iter().any(|c| self_.has_free_sym(&c, sym));
+            }
+        }
+
+        circuit.children().any(|c| self_.has_free_sym(&c, sym))
     }
 }
 
 /// uses global cache
-pub fn has_sym_impl(circuit: &CircuitRc, sym: &Symbol, by_ident: bool) -> bool {
-    HAS_SYM.with(|has_sym| has_sym.borrow_mut().has_sym(circuit, sym, by_ident))
-}
-pub fn has_sym(circuit: &CircuitRc, sym: &Symbol) -> bool {
-    has_sym_impl(circuit, sym, false)
-}
-pub fn has_sym_ident(circuit: &CircuitRc, sym: &Symbol) -> bool {
-    has_sym_impl(circuit, sym, true)
+pub fn has_free_sym(circuit: &CircuitRc, sym: &Symbol) -> bool {
+    HAS_FREE_SYM.with(|has_free_sym| has_free_sym.borrow_mut().has_free_sym(circuit, sym))
 }
 
-pub fn check_valid_sub(circuit: CircuitRc, sym: &Symbol, sub_for_sym: CircuitRc) -> Result<()> {
-    let hashs = [
-        circuit.info().hash,
-        sym.info().hash,
-        sub_for_sym.info().hash,
-    ];
-    if IS_VALID_SUB.with(|is_valid| is_valid.borrow().contains(&hashs)) {
-        return Ok(());
+// adhoc anyfound
+// TODO: do this batchedly over symbols while still caching individually?
+#[derive(Clone, Debug)]
+pub struct CheckValidSub {
+    cache: cached::UnboundCache<(HashBytes, HashBytes, HashBytes), ()>,
+}
+impl Default for CheckValidSub {
+    fn default() -> Self {
+        Self {
+            cache: cached::UnboundCache::new(),
+        }
     }
-
-    let ok = || {
-        IS_VALID_SUB.with(|is_valid| is_valid.borrow_mut().insert(hashs));
-    };
-
-    // we could do more caching I guess, but this seems fine.
-    let free: HashSet<_> = all_free(sub_for_sym.clone()).into_iter().collect();
-    let free_idents: HashSet<_> = free.iter().map(|x| x.ident()).collect();
-
-    visit_circuit(circuit, |x| {
-        if let Some(m) = x.as_module() {
-            if has_sym_ident(&m.spec.circuit, sym) {
-                for (arg_spec, node) in m.spec.arg_specs.iter().zip(&m.nodes) {
-                    if arg_spec.symbol.ident() == sym.ident() && node != &sub_for_sym {
-                        bail!(
-                            SubstitutionError::BindsSameSymbolToDifferentValueThanNestedModule {
-                                circ: sub_for_sym.clone(),
-                                sym: sym.clone(),
-                                nested_module: m.clone(),
-                                different_circ: node.clone()
-                            }
-                        );
-                    }
-                    if free_idents.contains(&arg_spec.symbol.ident()) {
-                        bail!(SubstitutionError::CircuitHasFreeSymBoundByNestedModule {
-                            circ: sub_for_sym.clone(),
-                            sym: sym.clone(),
-                            nested_module: m.clone(),
-                            free_sym: arg_spec.symbol.clone()
-                        });
+}
+impl CheckValidSub {
+    #[apply(cached_method)]
+    #[self_id(self_)]
+    #[key((circuit.info().hash, sym.info().hash, sub_for_sym.info().hash))]
+    #[use_try]
+    #[cache_expr(cache)]
+    pub fn check_valid_sub(
+        &mut self,
+        circuit: &CircuitRc,
+        sym: &Symbol,
+        sub_for_sym: &CircuitRc,
+    ) -> Result<()> {
+        if let Some(sym_here) = circuit.as_symbol() {
+            if sym_here != sym && sym_here.ident() == sym.ident() {
+                bail!(
+                    SubstitutionError::FoundNEQFreeSymbolWithSameIdentification {
+                        sym: sym.clone(),
+                        matching_sym: sym_here.clone()
                     }
+                )
+            }
+        }
+
+        circuit
+            .non_free_children() // do this first so we error at inner most module!
+            .map(|c| self_.check_valid_sub(&c, sym, sub_for_sym))
+            .collect::<Result<()>>()?;
+
+        if let Some(m) = circuit.as_module() {
+            if m.spec
+                .arg_specs
+                .iter()
+                .any(|arg_spec| &arg_spec.symbol == sym)
+            {
+                // skip spec circuit if bound
+                return m
+                    .nodes
+                    .iter()
+                    .map(|c| self_.check_valid_sub(&c, sym, sub_for_sym))
+                    .collect();
+            }
+            // otherwise, we might sub into the spec circuit
+            if has_free_sym(&m.spec.circuit, sym) {
+                let used_args = are_args_used(sub_for_sym, &m.spec.arg_specs);
+
+                if used_args.iter().any(|x| *x) {
+                    bail!(SubstitutionError::CircuitHasFreeSymsBoundByNestedModule {
+                        circ: sub_for_sym.clone(),
+                        sym: sym.clone(),
+                        nested_module: m.clone(),
+                        bound_by_nested: m
+                            .spec
+                            .arg_specs
+                            .iter()
+                            .zip(used_args)
+                            .filter_map(|(arg_spec, used)| used.then(|| arg_spec.symbol.clone()))
+                            .collect()
+                    })
                 }
             }
         }
+
         Ok(())
-    })?;
-    ok();
-    Ok(())
+    }
 }
 
+pub fn check_valid_sub(circuit: &CircuitRc, sym: &Symbol, sub_for_sym: &CircuitRc) -> Result<()> {
+    CHECK_VALID_SUB.with(|check_valid_sub| {
+        check_valid_sub
+            .borrow_mut()
+            .check_valid_sub(circuit, sym, sub_for_sym)
+    })
+}
+
+const SAME_IDENT_P1 :&str = "Substituting for a symbol when there is a not-equal free symbol with the same name and uuid is not allowed (aka, a near miss).";
+const SAME_IDENT_P2: &str = "This is caused by having a different shape (or different named axes).";
+
 #[apply(python_error_exception)]
 #[base_error_name(Substitution)]
 #[base_exception(PyValueError)]
 #[derive(Error, Debug, Clone)]
 pub enum SubstitutionError {
-    #[error("subbing circ={circ:?} for sym={sym:?} inconsistent with nested_module={nested_module:?} which binds equivalent sym to different_circ={different_circ:?} ({e_name})")]
-    BindsSameSymbolToDifferentValueThanNestedModule {
+    #[error("subbing circ={circ:?} for sym={sym:?} results in nested_module={nested_module:?} bound_by_nested={bound_by_nested:?} {} ({e_name})",
+        "inside of circ (aka higher order function)")]
+    CircuitHasFreeSymsBoundByNestedModule {
         circ: CircuitRc,
         sym: Symbol,
         nested_module: Module,
-        different_circ: CircuitRc,
+        bound_by_nested: Vec<Symbol>,
     },
 
-    #[error("subbing circ={circ:?} for sym={sym:?} results in nested_module={nested_module:?} replacing free_sym={free_sym:?} {} ({e_name})",
-        "inside of circ (aka higher order function)\nNote: it may be equivalent ignoring shape\n")]
-    CircuitHasFreeSymBoundByNestedModule {
-        circ: CircuitRc,
-        sym: Symbol,
-        nested_module: Module,
-        free_sym: Symbol,
-    },
-
-    #[error("symbols={symbols:?} nodes={nodes:?} idxs={idxs:?} at these idxs, symbols are equivalent, but have different nodes {} ({e_name})",
-        "(equivalence ignores shape)")]
-    BindsEquivalentSymbolsToDifferentNodes {
-        symbols: Vec<Symbol>,
-        nodes: Vec<CircuitRc>,
-        idxs: Vec<usize>,
-    },
+    #[error(
+        "sym={sym:?} matching_sym={matching_sym:?}\n{SAME_IDENT_P1}\n{SAME_IDENT_P2}\n({e_name})"
+    )]
+    FoundNEQFreeSymbolWithSameIdentification { sym: Symbol, matching_sym: Symbol },
 }
 
 use std::cell::RefCell;
 // TODO: maybe cache check_unique_arg_names here?
+// TODO: avoid caching arrays forever!
+// TODO: make modules much faster by avoiding exponential time blow up!
 thread_local! {
-    static MODULE_EXPANSIONS: RefCell<HashMap<(ModuleSpecHashable, Vec<HashBytes>, Option<String>), Weak<Circuit>>> =
-        RefCell::new(HashMap::default());
-    static MODULE_EXPANSIONS_SHAPE: RefCell<HashMap<(ModuleSpecHashable, Vec<Shape>), (Weak<Circuit>, Vec<HashBytes>)>> =
-        RefCell::new(HashMap::default());
-    static MODULE_EXPANSIONS_AXIS_MAPS: RefCell<HashMap<(ModuleSpecHashable, Vec<Shape>), Vec<Vec<Option<usize>>> >> =
-        RefCell::new(HashMap::default());
-    static IS_VALID_SUB: RefCell<HashSet<[HashBytes; 3]>> =
-        RefCell::new(Default::default());
-    static HAS_SYM: RefCell<HasSym>= RefCell::new(Default::default())
+    static MODULE_EXPANSIONS: RefCell<
+        HashMap<
+            (ModuleSpecHashable, Vec<HashBytes>, Option<String>),
+            CircuitRc, // Weak<Circuit>
+        >,
+    > = RefCell::new(HashMap::default());
+    static MODULE_EXPANSIONS_SHAPE: RefCell<
+        HashMap<
+            (ModuleSpecHashable, Vec<Shape>),
+            (
+                CircuitRc, // Weak<Circuit>
+                Vec<HashBytes>,
+            ),
+        >,
+    > = RefCell::new(HashMap::default());
+    static MODULE_EXPANSIONS_AXIS_MAPS: RefCell<
+        HashMap<(ModuleSpecHashable, Vec<Shape>), Vec<Vec<Option<usize>>>>,
+    > = RefCell::new(HashMap::default());
+    static CHECK_VALID_SUB: RefCell<CheckValidSub> = RefCell::new(Default::default());
+    static HAS_FREE_SYM: RefCell<HasFreeSym> = RefCell::new(Default::default());
 }
