@@ -43,7 +43,7 @@ pub struct Parser {
     #[pyo3(get, set)]
     pub on_repeat_check_info_same: bool,
     #[pyo3(get, set)]
-    pub module_check_all_args_present: bool,
+    pub module_check_all_inputs_used: bool,
     #[pyo3(get, set)]
     pub module_check_unique_arg_names: bool,
 }
@@ -56,7 +56,7 @@ impl Default for Parser {
             tensors_as_random_device_dtype: TorchDeviceDtypeOp::NONE,
             allow_hash_with_random: false,
             on_repeat_check_info_same: true,
-            module_check_all_args_present: true,
+            module_check_all_inputs_used: true,
             module_check_unique_arg_names: false,
         }
     }
@@ -105,7 +105,7 @@ impl Parser {
         tensors_as_random_device_dtype = "Parser::default().tensors_as_random_device_dtype",
         allow_hash_with_random = "Parser::default().allow_hash_with_random",
         on_repeat_check_info_same = "Parser::default().on_repeat_check_info_same",
-        module_check_all_args_present = "Parser::default().module_check_all_args_present",
+        module_check_all_inputs_used = "Parser::default().module_check_all_inputs_used",
         module_check_unique_arg_names = "Parser::default().module_check_unique_arg_names"
     )]
     #[new]
@@ -116,7 +116,7 @@ impl Parser {
         tensors_as_random_device_dtype: TorchDeviceDtypeOp,
         allow_hash_with_random: bool,
         on_repeat_check_info_same: bool,
-        module_check_all_args_present: bool,
+        module_check_all_inputs_used: bool,
         module_check_unique_arg_names: bool,
     ) -> Result<Self> {
         Ok(Self {
@@ -128,7 +128,7 @@ impl Parser {
             tensors_as_random_device_dtype,
             allow_hash_with_random,
             on_repeat_check_info_same,
-            module_check_all_args_present,
+            module_check_all_inputs_used,
             module_check_unique_arg_names,
         })
     }
@@ -899,7 +899,7 @@ impl Parser {
                     ModuleSpec::new(
                         spec_circuit,
                         arg_specs,
-                        self.module_check_all_args_present,
+                        self.module_check_all_inputs_used,
                         self.module_check_unique_arg_names,
                     )
                     .context("spec construction failed in parse")?,
