@@ -799,6 +799,7 @@ impl PrintOptions {
         fn recurse(
             circ: &Circuit,
             depth: usize,
+            is_only_child: bool,
             result: &mut String,
             seen_hashes: &mut HashMap<HashBytes, (String, String)>,
             runnning_serial_number: &mut usize,
@@ -810,7 +811,11 @@ impl PrintOptions {
             parent_info: Option<(&Circuit, usize)>,
             mut node_comments: Vec<String>,
         ) -> Result<()> {
-            result.push_str(&last_child_arrows(is_last_child, true, selfy.arrows));
+            result.push_str(&last_child_arrows(
+                is_last_child,
+                !(is_only_child && selfy.only_child_below),
+                selfy.arrows,
+            ));
             let mut used_color = false;
             if let Some(colorer) = &selfy.colorer {
                 if let Some(color) = colorer
@@ -925,6 +930,7 @@ impl PrintOptions {
                     recurse(
                         child,
                         depth + 1,
+                        circ.children().count() == 1,
                         result,
                         seen_hashes,
                         runnning_serial_number,
@@ -949,6 +955,7 @@ impl PrintOptions {
                 recurse(
                     &circ.children().next().unwrap(),
                     depth,
+                    circ.children().count() == 1,
                     result,
                     seen_hashes,
                     runnning_serial_number,
@@ -976,6 +983,7 @@ impl PrintOptions {
                 recurse(
                     &child,
                     depth + 1,
+                    circ.children().count() == 1,
                     result,
                     seen_hashes,
                     runnning_serial_number,
@@ -997,6 +1005,7 @@ impl PrintOptions {
             recurse(
                 &**circuit,
                 0,
+                circuit.children().count() == 1,
                 &mut result,
                 &mut seen_hashes,
                 &mut runnning_serial_number,
@@ -1206,7 +1215,12 @@ pub fn last_child_arrows(last_child: &Vec<bool>, is_output: bool, arrows: bool)
             });
             result.push_str(ARROW);
         } else {
-            write!(result, "{} ", BAR).unwrap();
+            write!(
+                result,
+                "{} ",
+                if !last_child[depth - 1] { BAR } else { " " }
+            )
+            .unwrap();
         }
     }
     result
