@@ -84,11 +84,11 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
     if input_specs.is_empty() {
         bail!(BatchError::AxisOriginatesTooHigh {});
     }
-    let check_all_args_present = false; // this really should be true - get_axis_leaves doesn't get leaves!
+    let check_all_inputs_used = false; // this really should be true - get_axis_leaves doesn't get leaves!
     let module_spec = ModuleSpec::new_extract(
         circuit.clone(),
         input_specs.clone(),
-        check_all_args_present,
+        check_all_inputs_used,
         false,
     )
     .unwrap();
