@@ -2,8 +2,8 @@ use std::iter::zip;
 
 use anyhow::{anyhow, bail, Context, Result};
 use circuit_base::{
-    circuit_utils::count_nodes, deep_map_op_context, module::has_sym, CircuitNode, CircuitRc,
-    Concat, Module, ModuleArgSpec, ModuleSpec, Rearrange,
+    circuit_utils::count_nodes, deep_map_op_context, CircuitNode, CircuitRc, Concat, Module,
+    ModuleArgSpec, ModuleSpec, Rearrange,
 };
 use pyo3::prelude::*;
 use rr_util::{
@@ -56,19 +56,22 @@ pub fn module_remove_unused_inputs_op(
     m: &Module,
     use_elim_no_input_module: bool,
 ) -> Result<Option<CircuitRc>> {
-    let (nodes, arg_specs): (Vec<_>, Vec<_>) = m
-        .arg_items()
-        .into_iter()
-        .filter(|(_, ModuleArgSpec { symbol, .. })| has_sym(&m.spec.circuit, symbol))
-        .unzip();
+    let used_inputs = m.spec.are_args_used();
 
-    if nodes.len() == m.nodes.len() {
+    if used_inputs.iter().all(|x| *x) {
         if use_elim_no_input_module {
             return Ok(elim_no_input_module(m));
         }
         return Ok(None);
     }
 
+    let (nodes, arg_specs) = m
+        .arg_items()
+        .into_iter()
+        .zip(used_inputs)
+        .filter_map(|(x, used)| used.then_some(x))
+        .unzip();
+
     let m_out = Module::new(
         nodes,
         ModuleSpec {
@@ -115,7 +118,7 @@ pub fn module_remove_unused_inputs(
 #[pyfunction(
     prefix_to_strip = "None",
     module_name = "None",
-    check_all_args_present = "true",
+    check_all_inputs_used = "true",
     check_unique_arg_names = "true"
 )]
 pub fn extract_rewrite_raw(
@@ -123,13 +126,13 @@ pub fn extract_rewrite_raw(
     input_specs: Vec<(CircuitRc, ModuleArgSpec)>,
     prefix_to_strip: Option<String>,
     module_name: Option<String>,
-    check_all_args_present: bool,
+    check_all_inputs_used: bool,
     check_unique_arg_names: bool,
 ) -> Result<Module> {
     let mut spec = ModuleSpec::new_extract(
         circuit,
         input_specs.clone(),
-        check_all_args_present,
+        check_all_inputs_used,
         check_unique_arg_names,
     )?;
     if let Some(pref) = &prefix_to_strip {
