@@ -905,7 +905,7 @@ fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationConte
             Instruction::Compute(outermost_k, v) => match &***v {
                 Circuit::Module(mn) => {
                     mn.spec
-                        .check_all_args_present()
+                        .check_all_inputs_used()
                         .expect("unused args removed at start of circuit_to_schedule!");
 
                     let original_children: Vec<usize> =
