@@ -15,7 +15,11 @@ use circuit_base::{
     CircuitNode, CircuitRc,
 };
 use macro_rules_attribute::apply;
-use pyo3::{exceptions::PyValueError, prelude::*, AsPyPointer};
+use pyo3::{
+    exceptions::{PyTypeError, PyValueError},
+    prelude::*,
+    AsPyPointer,
+};
 use rr_util::{
     eq_by_big_hash::EqByBigHash,
     impl_both_by_big_hash, python_error_exception, setup_callable, simple_default, simple_from,
@@ -1020,6 +1024,9 @@ impl IterativeMatcher {
     fn __rand__(&self, other: IterativeMatcherRc) -> Self {
         arc_unwrap_or_clone(other.0).and(self.crc())
     }
+    fn __bool__(&self) -> PyResult<bool> {
+        PyResult::Err(PyTypeError::new_err("IterativeMatcher was coerced to a boolean. Did you mean to use & or | instead of \"and\" or \"or\"?"))
+    }
 
     // TODO: write flatten/simplify method if we want the extra speed + niceness!
 }
