@@ -4,16 +4,20 @@ use std::{
     iter::{once, zip},
 };
 
-use anyhow::Result;
+use anyhow::{Error, Result};
 use circuit_base::{
     clicolor,
     print::{color, last_child_arrows, CliColor, PrintOptions},
     CircuitNode, CircuitNodeUnion, CircuitRc,
 };
 use itertools::Itertools;
+use once_cell::unsync::Lazy;
 use pyo3::prelude::*;
-use rr_util::python_println;
-use rustc_hash::FxHashSet as HashSet;
+use rr_util::{
+    python_println,
+    util::{indent, indent_butfirst},
+};
+use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 
 use crate::{
     iterative_matcher::{per_child, ChainItem, IterativeMatcher},
@@ -21,10 +25,53 @@ use crate::{
     IterateMatchResults, IterativeMatcherData, IterativeMatcherRc,
 };
 
+const MAX_SELF_SINGLE_LINE_WIDTH: Lazy<usize> = Lazy::new(|| {
+    std::env::var("MATCHER_PRINT_SELF_WIDTH")
+        .ok()
+        .and_then(|x| x.parse::<usize>().ok())
+        .unwrap_or(80)
+});
+const MATCHER_PRINT_INDENT_WIDTH: usize = 4;
+
+fn indented_if_multiline(x: &str) -> String {
+    if x.contains("\n") {
+        return format!("\n{}\n", indent(x.to_owned(), MATCHER_PRINT_INDENT_WIDTH));
+    }
+    x.to_owned()
+}
+fn maybe_indented_list(
+    strings: Vec<String>,
+    sep: &str,
+    sep_after_line: bool,
+    surrounding_newlines: bool,
+) -> String {
+    let total_length: usize = strings.iter().map(|s| s.len() + 3).sum();
+    if total_length < *MAX_SELF_SINGLE_LINE_WIDTH && !strings.iter().any(|s| s.contains("\n")) {
+        return strings.iter().join(sep);
+    }
+    let sep: String = if sep_after_line {
+        format!("{}\n", sep)
+    } else {
+        "\n".to_owned()
+    };
+    let mut result = indent(
+        strings
+            .iter()
+            .map(|s| indent_butfirst(s.clone(), MATCHER_PRINT_INDENT_WIDTH))
+            .join(&sep),
+        MATCHER_PRINT_INDENT_WIDTH,
+    );
+    if surrounding_newlines {
+        result = format!("\n{}\n", result);
+    }
+    result
+}
+
 impl Display for IterativeMatcher {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         let s = match self.data() {
             IterativeMatcherData::Match(matcher) => format!("{}", matcher),
+            IterativeMatcherData::Raw(_) => "RawIterativeMatcher".to_owned(),
             IterativeMatcherData::Term(term) => {
                 if *term {
                     "Term".to_owned()
@@ -32,38 +79,81 @@ impl Display for IterativeMatcher {
                     "term next".to_owned()
                 }
             }
-            IterativeMatcherData::Filter(filter) => format!(
-                "Filter{{ depth_range {}:{}, matcher {}, term_at {}, term_if_matches {}, depth {} }}",
-                op_debug(&filter.start_depth),
-                op_debug(&filter.end_depth),
-                &filter.iterative_matcher.0,
-                &filter.term_early_at.0,
-                filter.term_if_matches,
-                filter.depth
-            ),
+            IterativeMatcherData::PyFunc(pyobject) => {
+                format!(
+                    "PyIterativeMatcher({})",
+                    indented_if_multiline(&format!("{}", pyobject))
+                )
+            }
+            IterativeMatcherData::Filter(filter) => {
+                let mut entries: Vec<String> = vec![];
+                if filter.start_depth.is_some() || filter.end_depth.is_some() {
+                    entries.push(format!(
+                        "depth_range: {}:{}",
+                        op_debug(&filter.start_depth),
+                        op_debug(&filter.end_depth)
+                    ));
+                }
+                if &**filter.iterative_matcher != &Matcher::true_matcher().into() {
+                    entries.push(format!("matcher: {}", filter.iterative_matcher.0));
+                }
+                if &**filter.term_early_at != &Matcher::false_matcher() {
+                    entries.push(format!("term_at: {}", filter.term_early_at.0));
+                }
+                if filter.term_if_matches {
+                    entries.push("term_if_matches: true".to_owned());
+                }
+                if filter.depth != 0 {
+                    entries.push(format!("cur_depth:{}", filter.depth));
+                }
+                format!(
+                    "Filter{{{}}}",
+                    maybe_indented_list(entries, ", ", false, true)
+                )
+            }
             IterativeMatcherData::Children(childmatcher) => format!(
-                "Children({}, {})",
-                &childmatcher.iterative_matcher.0,
+                "Children({}, [{}])",
+                indented_if_multiline(&format!("{}", &childmatcher.iterative_matcher.0)),
                 childmatcher
                     .child_numbers
                     .iter()
                     .map(|x| x.to_string())
                     .join(","),
             ),
-            IterativeMatcherData::SpecCircuit(matcher) => format!("SpecCircuit({})", &matcher.0),
-            IterativeMatcherData::NoModuleSpec(matcher) => format!("NoModuleSpec({})", &matcher.0),
+            IterativeMatcherData::SpecCircuit(matcher) => format!(
+                "SpecCircuit({})",
+                indented_if_multiline(&format!("{}", &matcher.0))
+            ),
+            IterativeMatcherData::NoModuleSpec(matcher) => format!(
+                "NoModuleSpec({})",
+                indented_if_multiline(&format!("{}", &matcher.0))
+            ),
             IterativeMatcherData::Chains(chains) => {
                 if chains.len() == 1 {
                     repr_chain_item(chains.iter().next().unwrap())
                 } else {
-                    format!("Any({})", chains.iter().map(repr_chain_item).join(", "))
+                    format!(
+                        "Any({})",
+                        maybe_indented_list(
+                            chains.iter().map(repr_chain_item).collect(),
+                            ", ",
+                            false,
+                            true
+                        )
+                    )
                 }
             }
             IterativeMatcherData::All(matchers) => {
-                format!("All({})", matchers.iter().map(|x| x.to_string()).join(", "))
+                format!(
+                    "All({})",
+                    maybe_indented_list(
+                        matchers.iter().map(|x| x.to_string()).collect(),
+                        ", ",
+                        false,
+                        true
+                    )
+                )
             }
-            IterativeMatcherData::PyFunc(pyobject) => format!("PyIterativeMatcher {}", pyobject),
-            IterativeMatcherData::Raw(_) => "".to_owned(),
         };
         f.write_str(&s)
     }
@@ -77,10 +167,21 @@ impl IterativeMatcher {
 }
 
 fn repr_chain_item(chain_item: &ChainItem) -> String {
-    once(chain_item.first.clone())
-        .chain(chain_item.rest.iter().cloned())
-        .map(|x| format!("{}", &x.0))
-        .join(" -> ")
+    if chain_item.rest.is_empty() {
+        return format!("{}", &chain_item.first.0);
+    }
+    format!(
+        "[{}]",
+        maybe_indented_list(
+            once(chain_item.first.clone())
+                .chain(chain_item.rest.iter().cloned())
+                .map(|x| format!("{}", &x.0))
+                .collect(),
+            " -> ",
+            true,
+            true
+        )
+    )
 }
 
 fn op_debug<T: std::fmt::Debug>(x: &Option<T>) -> String {
@@ -97,7 +198,10 @@ where
     if x.len() == 1 {
         return f(x.iter().next().unwrap());
     }
-    format!("{{{}}}", x.iter().map(f).join(", "))
+    format!(
+        "{{{}}}",
+        maybe_indented_list(x.iter().map(f).collect(), ", ", false, true)
+    )
 }
 
 pub fn circuit_short_print(circuit: &CircuitRc) -> String {
@@ -136,11 +240,21 @@ impl Display for Matcher {
                 MatcherData::Not(matcher) => format!("Not {}", &matcher.0),
                 MatcherData::Any(any) => format!(
                     "Any({})",
-                    any.iter().map(|x| format!("{}", &x.0)).join(", ")
+                    maybe_indented_list(
+                        any.iter().map(|x| format!("{}", &x.0)).collect(),
+                        ", ",
+                        false,
+                        true
+                    ),
                 ),
                 MatcherData::All(all) => format!(
                     "All({})",
-                    all.iter().map(|x| format!("{}", &x.0)).join(", ")
+                    maybe_indented_list(
+                        all.iter().map(|x| format!("{}", &x.0)).collect(),
+                        ", ",
+                        false,
+                        true
+                    ),
                 ),
                 // MatcherData::Raw(RawMatcher),
                 _ => "".to_owned(),
@@ -277,3 +391,14 @@ pub fn append_matchers_to_names(
     }
     recurse(circuit, Some(matcher), discard_old_name)
 }
+// Filter{
+//     m: Filter{
+//         m: Chain(
+//             Filter{ m: Always, term_if_matches true }
+//             ->
+//             "m1_pos0"
+//         ),
+//         term: {"a0.out_pos0", "a1.out_pos0", "a2.out_pos0", "m0_pos0", "m2_pos0"}
+//     } -> "a0.out_pos0_head_1_sum",
+//     term: {"a0.out_pos0", "a1.out_pos0", "a2.out_pos0", "m0_pos0", "m1_pos0", "m2_pos0"},
+// }
