@@ -753,21 +753,16 @@ impl Expander {
 
         let new_children = circuit
             .children()
+            .zip(extra_replacements.per_child(&circuit))
             .zip(new_matchers_per.clone())
-            .map(|(c, new_matchers)| self_.batch_impl(c, new_matchers, extra_replacements))
+            .map(|((c, rep), new_matchers)| self_.batch_impl(c, new_matchers, &rep))
             .collect::<Result<_>>()?;
 
         let expanded = expand_node(circuit.clone(), &new_children, |c, rep, child_idx| {
             self_.batch_impl(
                 c,
                 new_matchers_per[child_idx].clone(),
-                &ReplaceMapRc::new(
-                    extra_replacements
-                        .iter()
-                        .map(|(a, b)| (a.clone(), b.clone()))
-                        .chain(rep)
-                        .collect(),
-                ),
+                &extra_replacements.extend_into(rep),
             )
         })?;
         let expanded = if expanded != circuit {
