@@ -1,5 +1,4 @@
 use std::{
-    collections::{HashMap, HashSet},
     fmt::Debug,
     iter::zip,
     sync::{Arc, Mutex},
@@ -26,6 +25,7 @@ use rr_util::{
     tu8v,
     util::{flip_op_result, HashBytes},
 };
+use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use smallvec::SmallVec as Sv;
 use thiserror::Error;
 
@@ -209,7 +209,7 @@ impl RandomSampleSpec {
         .into_init()
     }
 
-    pub fn indices_cache(&self) -> HashMap<HashBytes, Tensor> {
+    pub fn indices_cache(&self) -> ::std::collections::HashMap<HashBytes, Tensor> {
         self.indices_cache.lock().unwrap().get_store().clone()
     }
 }
@@ -412,7 +412,7 @@ impl RunDiscreteVarAllSpec {
     #[staticmethod]
     #[args(circuits = "*")]
     fn create_full_from_circuits(circuits: Vec<CircuitRc>) -> Result<Self> {
-        let mut discrete_vars_groups: HashSet<CircuitRc> = HashSet::new();
+        let mut discrete_vars_groups: HashSet<CircuitRc> = HashSet::default();
         let mut add_all_groups = |circ: CircuitRc| {
             visit_circuit(circ, |sub| {
                 if let Some(var) = sub.as_discrete_var() {
