@@ -1,9 +1,14 @@
-use std::collections::{btree_map::Entry, BTreeMap};
+use std::{
+    collections::{btree_map::Entry, BTreeMap, BTreeSet},
+    iter,
+    ops::Deref,
+    sync::Arc,
+};
 
 use anyhow::{anyhow, bail, Context, Result};
 use circuit_base::{
-    expand_node::{expand_node, ReplaceMapRc},
-    module::has_sym,
+    expand_node::{expand_node, map_replace_expand, ReplaceMapRc},
+    module::are_args_used,
     CircuitNode, CircuitRc, Module, ModuleArgSpec, ModuleSpec, Rearrange, Symbol,
 };
 use circuit_rewrites::{
@@ -23,7 +28,7 @@ use rr_util::{
     fn_struct, impl_eq_by_big_hash, pycall, python_error_exception,
     rearrange_spec::{OpSize, RearrangeSpec, RearrangeSpecError},
     sv, tu8v,
-    util::{DimNumMaker, HashBytes},
+    util::{arc_unwrap_or_clone, DimNumMaker, HashBytes},
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use thiserror::Error;
@@ -31,7 +36,7 @@ use thiserror::Error;
 #[pyfunction(
     prefix_to_strip = "None",
     module_name = "None",
-    check_all_args_present = "true",
+    check_all_inputs_used = "true",
     check_unique_arg_names = "true",
     circuit_to_arg_spec = "None"
 )]
@@ -40,7 +45,7 @@ pub fn extract_rewrite(
     matcher: IterativeMatcherRc,
     prefix_to_strip: Option<String>,
     module_name: Option<String>,
-    check_all_args_present: bool,
+    check_all_inputs_used: bool,
     check_unique_arg_names: bool,
     circuit_to_arg_spec: Option<PyObject>,
 ) -> Result<Module> {
@@ -62,7 +67,7 @@ pub fn extract_rewrite(
         specs,
         prefix_to_strip,
         module_name,
-        check_all_args_present,
+        check_all_inputs_used,
         check_unique_arg_names,
     )
 }
@@ -217,7 +222,7 @@ pub fn fuse_modules_impl(
         .into_iter()
         .rev(); // reverse again
 
-    let out: Vec<_> = new_nodes
+    let out = new_nodes
         .into_iter()
         .zip(cumulative_batch_shapes)
         .zip(orig_modules)
@@ -295,25 +300,6 @@ pub fn fuse_modules_impl(
         .flatten()
         .collect();
 
-    // resolve all references
-    let out = out
-        .into_iter()
-        .scan(
-            HashMap::<_, CircuitRc>::default(),
-            |sym_map, (circ, arg_spec)| {
-                let circ = if let Some(circ) = sym_map.get(&arg_spec.symbol) {
-                    circ.clone()
-                } else {
-                    circ
-                };
-
-                sym_map.insert(arg_spec.symbol.clone(), circ.clone());
-
-                Some((circ, arg_spec))
-            },
-        )
-        .collect();
-
     Ok(out)
 }
 
@@ -334,6 +320,10 @@ impl Default for NestedModuleItems {
 
 impl NestedModuleItems {
     fn new(modules: Vec<Module>, new_nodes: Vec<Vec<CircuitRc>>) -> Result<Self> {
+        for (self_m, new) in modules.iter().zip(&new_nodes) {
+            assert_eq!(self_m.nodes.len(), new.len());
+        }
+
         let flat = fuse_modules_impl(&modules, &new_nodes)?;
         let flat_sym_to_arg = flat
             .iter()
@@ -440,7 +430,6 @@ pub struct ModulePusher {
         ),
         (CircuitRc, HashSet<CircuitRc>),
     >,
-    replace_expand_cache: cached::UnboundCache<(HashBytes, HashBytes), CircuitRc>,
 }
 
 impl Default for ModulePusher {
@@ -453,7 +442,6 @@ impl Default for ModulePusher {
             namer: Default::default(),
             any_found: Default::default(),
             cache: cached::UnboundCache::new(),
-            replace_expand_cache: cached::UnboundCache::new(),
         }
     }
 }
@@ -578,21 +566,6 @@ impl ModulePusher {
         items: &NestedModuleItems,
         extra_replacements: &ReplaceMapRc,
     ) -> Result<(CircuitRc, HashSet<CircuitRc>)> {
-        if let Some(arg) = circuit
-            .as_symbol()
-            .and_then(|sym| items.flat_sym_to_arg.get(sym))
-        {
-            if !self_.bind_encountered_symbols {
-                bail!(PushDownModuleError::PushPastPreviouslyBoundSymbol {
-                    symbol: circuit.as_symbol().unwrap().clone()
-                })
-            }
-            return Ok((arg.clone(), Default::default()));
-        }
-        if let Some(out) = extra_replacements.get(&circuit) {
-            return Ok((out.clone(), Default::default()));
-        }
-
         let (new_get, found_get) = get
             .match_iterate(circuit.clone())?
             .unwrap_or_same(get.clone());
@@ -620,6 +593,21 @@ impl ModulePusher {
             return Ok((out, out_set));
         }
 
+        if let Some(arg) = circuit
+            .as_symbol()
+            .and_then(|sym| items.flat_sym_to_arg.get(sym))
+        {
+            if !self_.bind_encountered_symbols {
+                bail!(PushDownModuleError::PushPastPreviouslyBoundSymbol {
+                    symbol: circuit.as_symbol().unwrap().clone()
+                })
+            }
+            return Ok((arg.clone(), Default::default()));
+        }
+        if let Some(out) = extra_replacements.get(&circuit) {
+            return Ok((out.clone(), Default::default()));
+        }
+
         let new_get = new_get.per_child_with_term(circuit.num_children());
 
         let (new_skip_module, found_skip_module) = skip_module
@@ -645,6 +633,7 @@ impl ModulePusher {
             let node_get_skip: Vec<_> = new_iter.collect();
             assert_eq!(node_get_skip.len(), m.nodes.len());
 
+            // intentionally don't use extra_replacements per child
             let new_nodes = m
                 .nodes
                 .clone()
@@ -662,29 +651,35 @@ impl ModulePusher {
                 extra_replacements,
             )?
         } else {
+            if let Some(m) = circuit.as_module() {
+                for arg_spec in &m.spec.arg_specs {
+                    if items.flat_sym_to_arg.contains_key(&arg_spec.symbol) {
+                        // we could add support for this case, but quite annoying
+                        bail!(PushDownModuleError::PushingPastModuleWhichOverridesSym {
+                            symbol: arg_spec.symbol.clone(),
+                            skipped_module: m.clone()
+                        });
+                    }
+                }
+            }
+
             let new_children = circuit
                 .children()
                 .into_iter()
                 .zip(new_get.clone())
                 .zip(new_skip_module.clone())
-                .map(|((c, get), skip)| rec(self_, c, get, skip, items, extra_replacements))
+                .zip(extra_replacements.per_child(&circuit))
+                .map(|(((c, get), skip), rep)| rec(self_, c, get, skip, items, &rep))
                 .collect::<Result<Vec<_>>>()?;
 
             expand_node(circuit.clone(), &new_children, |c, rep, child_idx| {
-                let rep = ReplaceMapRc::new(
-                    extra_replacements
-                        .iter()
-                        .map(|(a, b)| (a.clone(), b.clone()))
-                        .chain(rep)
-                        .collect(),
-                );
                 rec(
                     self_,
                     c,
                     new_get[child_idx].clone(),
                     new_skip_module[child_idx].clone(),
                     items,
-                    &rep,
+                    &extra_replacements.extend_into(rep),
                 )
             })
             // maybe we're supposed to just panic on rearrange rank errors?
@@ -718,7 +713,7 @@ impl ModulePusher {
         items: &NestedModuleItems,
         extra_replacements: &ReplaceMapRc,
     ) -> Result<CircuitRc> {
-        let finished_circuit = self.finish_replace_expand(circuit.clone(), extra_replacements)?;
+        let finished_circuit = map_replace_expand(circuit.clone(), extra_replacements)?;
         let rep_expand_extra_dims = finished_circuit.ndim() - circuit.ndim();
 
         let mut updated_extra_replacements = (**extra_replacements).clone();
@@ -761,7 +756,7 @@ impl ModulePusher {
             let (finished_circuit, new_flat) = if rep_expand_extra_dims > 0 {
                 // if we have extra dims, we have to pull the symbols and resub to sync up batching
                 let new_flat = update_flat(items.flat.clone(), rep_expand_extra_dims);
-                let new_finished_circuit = self.finish_replace_expand(
+                let new_finished_circuit = map_replace_expand(
                     circuit.clone(),
                     &ReplaceMapRc::new(updated_extra_replacements),
                 )?;
@@ -791,7 +786,7 @@ impl ModulePusher {
                         Some(update_flat(m.arg_items(), extra_dims_here))
                     })
                     .collect();
-                let new_finished_circuit = self.finish_replace_expand(
+                let new_finished_circuit = map_replace_expand(
                     circuit.clone(),
                     &ReplaceMapRc::new(updated_extra_replacements),
                 )?;
@@ -858,47 +853,65 @@ impl ModulePusher {
             },
         )
     }
+}
 
-    #[apply(cached_method)]
-    #[self_id(self_)]
-    #[key((circuit.info().hash,  extra_replacements.hash()))]
-    #[use_try]
-    #[cache_expr(replace_expand_cache)]
-    fn finish_replace_expand(
-        &mut self,
-        circuit: CircuitRc,
-        extra_replacements: &ReplaceMapRc,
-    ) -> Result<CircuitRc> {
-        if extra_replacements.is_empty() {
-            return Ok(circuit);
-        }
-        if let Some(out) = extra_replacements.get(&circuit) {
-            return Ok(out.clone());
-        }
-        let new_args = circuit
-            .children()
-            .map(|c| self_.finish_replace_expand(c, &extra_replacements))
-            .collect::<Result<_>>()?;
-        expand_node(circuit, &new_args, |c, rep, _child_idx| {
-            self_.finish_replace_expand(
-                c,
-                &ReplaceMapRc::new(
-                    extra_replacements
-                        .iter()
-                        .map(|(a, b)| (a.clone(), b.clone()))
-                        .chain(rep)
-                        .collect(),
-                ),
-            )
-        })
+#[derive(Clone, Debug)]
+pub struct SymbolSetRc {
+    set: Arc<BTreeSet<Symbol>>,
+    hash: HashBytes,
+}
+
+impl<'source> pyo3::FromPyObject<'source> for SymbolSetRc {
+    fn extract(from_py_obj: &'source pyo3::PyAny) -> pyo3::PyResult<Self> {
+        let from: BTreeSet<Symbol> = from_py_obj.extract()?;
+        Ok(Self::new(from))
     }
 }
 
+impl Default for SymbolSetRc {
+    fn default() -> Self {
+        Self::new(Default::default())
+    }
+}
+
+impl Deref for SymbolSetRc {
+    type Target = BTreeSet<Symbol>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.set
+    }
+}
+
+impl SymbolSetRc {
+    pub fn new(set: BTreeSet<Symbol>) -> Self {
+        let mut hasher = blake3::Hasher::new();
+        for v in set.iter() {
+            hasher.update(&v.hash());
+        }
+        Self {
+            set: Arc::new(set),
+            hash: hasher.finalize().into(),
+        }
+    }
+
+    pub fn into_map(self) -> <Self as Deref>::Target {
+        arc_unwrap_or_clone(self.set)
+    }
+}
+
+impl EqByBigHash for SymbolSetRc {
+    fn hash(&self) -> HashBytes {
+        self.hash
+    }
+}
+impl_eq_by_big_hash!(SymbolSetRc);
+
 struct SymbolExtractor {
-    cache:
-        cached::UnboundCache<HashBytes, (CircuitRc, BTreeMap<Symbol, (CircuitRc, ModuleArgSpec)>)>,
+    cache: cached::UnboundCache<
+        (HashBytes, HashBytes),
+        (CircuitRc, BTreeMap<Symbol, (CircuitRc, ModuleArgSpec)>),
+    >,
     conform_batch_if_needed: bool,
-    symbols: HashSet<Symbol>,
 }
 
 impl Default for SymbolExtractor {
@@ -906,7 +919,6 @@ impl Default for SymbolExtractor {
         Self {
             cache: cached::UnboundCache::new(),
             conform_batch_if_needed: false,
-            symbols: Default::default(),
         }
     }
 }
@@ -914,12 +926,13 @@ impl Default for SymbolExtractor {
 impl SymbolExtractor {
     #[apply(cached_method)]
     #[self_id(self_)]
-    #[key(circ.info().hash)]
+    #[key((circ.info().hash, symbols.hash()))]
     #[use_try]
     #[cache_expr(cache)]
     fn extract_symbols_rec(
         &mut self,
         circ: CircuitRc,
+        symbols: &SymbolSetRc,
     ) -> Result<(CircuitRc, BTreeMap<Symbol, (CircuitRc, ModuleArgSpec)>)> {
         let mut out_syms = BTreeMap::default();
         let add_sym_circ = |out_syms: &mut BTreeMap<_, (CircuitRc, ModuleArgSpec)>,
@@ -948,6 +961,11 @@ impl SymbolExtractor {
             Ok(())
         };
 
+        // do before we extract
+        // TODO: add flag for nested extract?
+        // let per_child_syms = symbols.per_child(&circ);
+
+        let mut freshly_bound_syms = HashSet::default(); // so we extract outer -> inner (left -> right)
         let circ = if let Some(m) = circ.as_module() {
             let new_mod = if self_.conform_batch_if_needed {
                 let batch = m
@@ -956,8 +974,7 @@ impl SymbolExtractor {
                     .zip(&m.spec.arg_specs)
                     .filter_map(|(node, arg_spec)| {
                         assert!(node.ndim() >= arg_spec.symbol.ndim());
-                        self_
-                            .symbols
+                        symbols
                             .contains(&arg_spec.symbol)
                             .then_some(node.ndim() - arg_spec.symbol.ndim())
                     })
@@ -972,7 +989,9 @@ impl SymbolExtractor {
                 .into_iter()
                 .zip(&m.spec.arg_specs)
                 .filter_map(|((node, arg_spec), orig_arg_spec)| {
-                    if !self_.symbols.contains(&orig_arg_spec.symbol) {
+                    if !symbols.contains(&orig_arg_spec.symbol)
+                        || freshly_bound_syms.contains(&orig_arg_spec.symbol)
+                    {
                         return Some(Ok((node, arg_spec)));
                     }
                     if node.ndim() > arg_spec.symbol.ndim() {
@@ -988,6 +1007,7 @@ impl SymbolExtractor {
                     if let Err(e) = add_sym_circ(&mut out_syms, arg_spec, node) {
                         return Some(Err(e));
                     }
+                    freshly_bound_syms.insert(orig_arg_spec.symbol.clone());
                     None
                 })
                 .collect::<Result<Vec<_>>>()?
@@ -1009,8 +1029,22 @@ impl SymbolExtractor {
             circ
         };
 
-        let new_circ = circ.map_children(|c| {
-            let (new, map) = self_.extract_symbols_rec(c)?;
+        let per_child_syms = if !freshly_bound_syms.is_empty() {
+            let mut removed = (**symbols).clone();
+            for bound in &freshly_bound_syms {
+                removed.remove(bound);
+            }
+
+            iter::once(SymbolSetRc::new(removed))
+                .chain(iter::repeat(symbols).cloned())
+                .take(circ.num_children())
+                .collect()
+        } else {
+            vec![symbols.clone(); circ.num_children()]
+        };
+
+        let new_circ = circ.map_children_enumerate(|i, c| {
+            let (new, map) = self_.extract_symbols_rec(c, &per_child_syms[i])?;
             for (_, (circ, arg_spec)) in map {
                 add_sym_circ(&mut out_syms, arg_spec, circ)?;
             }
@@ -1019,14 +1053,19 @@ impl SymbolExtractor {
 
         if let Some(m) = circ.as_module() {
             for (sym, (bound_input, _)) in &out_syms {
-                for arg_spec in &m.spec.arg_specs {
-                    if has_sym(bound_input, &arg_spec.symbol) {
-                        bail!(ExtractSymbolsError::HasBindingFromOuterModule {
-                            bound_input: bound_input.clone(),
-                            symbol: sym.clone(),
-                            outer_symbol: arg_spec.symbol.clone()
-                        })
-                    }
+                let used_args = are_args_used(bound_input, &m.spec.arg_specs);
+                if used_args.iter().any(|x| *x) {
+                    bail!(ExtractSymbolsError::HasBindingsFromOuterModule {
+                        bound_input: bound_input.clone(),
+                        symbol: sym.clone(),
+                        free_outer_symbols: m
+                            .spec
+                            .arg_specs
+                            .iter()
+                            .zip(used_args)
+                            .filter_map(|(arg_spec, used)| used.then(|| arg_spec.symbol.clone()))
+                            .collect(),
+                    })
                 }
             }
         }
@@ -1038,17 +1077,16 @@ impl SymbolExtractor {
 #[pyfunction(use_elim_no_input_modules = "true", conform_batch_if_needed = "false")]
 pub fn extract_symbols(
     circuit: CircuitRc,
-    symbols: HashSet<Symbol>,
+    symbols: SymbolSetRc,
     use_elim_no_input_modules: bool,
     conform_batch_if_needed: bool,
 ) -> Result<Module> {
     let name = circuit.name_cloned();
     let (extracted_circ, extracted_syms) = SymbolExtractor {
-        symbols,
         conform_batch_if_needed,
         ..Default::default()
     }
-    .extract_symbols_rec(circuit)?;
+    .extract_symbols_rec(circuit, &symbols)?;
 
     let extracted_circ = if use_elim_no_input_modules {
         SimpFnSubset::none()
@@ -1094,12 +1132,15 @@ pub fn extract_symbols_get(
         .collect::<Result<_>>()?;
     extract_symbols(
         circuit,
-        syms,
+        SymbolSetRc::new(syms),
         use_elim_no_input_modules,
         conform_batch_if_needed,
     )
 }
 
+const PREFIX_PAST: &str = "(Not currently supported) the encountered";
+const MIDDLE_PAST: &str = "was bound by a pushed module, but the";
+
 #[apply(python_error_exception)]
 #[base_error_name(PushDownModule)]
 #[base_exception(PyValueError)]
@@ -1107,6 +1148,12 @@ pub fn extract_symbols_get(
 pub enum PushDownModuleError {
     #[error("the encountered symbol={symbol:?} was bound by a pushed module, but bind_encountered_symbols is false ({e_name})")]
     PushPastPreviouslyBoundSymbol { symbol: Symbol },
+
+    #[error("{PREFIX_PAST} symbol={symbol:?} {MIDDLE_PAST} skipped_module={skipped_module:?} overrides this binding ({e_name})")]
+    PushingPastModuleWhichOverridesSym {
+        symbol: Symbol,
+        skipped_module: Module,
+    },
 }
 
 #[apply(python_error_exception)]
@@ -1145,10 +1192,10 @@ pub enum ExtractSymbolsError {
         new_bound: CircuitRc,
     },
 
-    #[error("the bound_input={bound_input:?} (for symbol={symbol:?}) contains outer_symbol={outer_symbol:?} which is bound by an outer module ({e_name})")]
-    HasBindingFromOuterModule {
+    #[error("the bound_input={bound_input:?} (for symbol={symbol:?}) contains free_outer_symbols={free_outer_symbols:?} which is bound by an outer module ({e_name})")]
+    HasBindingsFromOuterModule {
         bound_input: CircuitRc,
         symbol: Symbol,
-        outer_symbol: Symbol,
+        free_outer_symbols: Vec<Symbol>,
     },
 }
