@@ -22,11 +22,19 @@ dtype_from_string = {
     "float32": torch.float32,
     "float64": torch.float64,
     "float16": torch.float16,
+    "bfloat16": torch.bfloat16,
     "int64": torch.int64,
     "int32": torch.int32,
     "int16": torch.int16,
     "int8": torch.int8,
+    "bool": torch.bool,
+    "uint8": torch.uint8,
 }
+dtype_to_string = {v: k for k, v in dtype_from_string.items()}
+
+
+def maybe_dtype_to_maybe_string(x):
+    return dtype_to_string.get(x)
 
 
 def scalar_to_tensor(scalar, shape, device_dtype):
@@ -40,7 +48,12 @@ def tensor_scale(tensor):
 
 
 def cast_tensor(tensor, device_dtype):
-    return tensor.to(device=device_dtype.device, dtype=dtype_from_string[device_dtype.dtype])
+    kwargs = {}
+    if device_dtype.device is not None:
+        kwargs["device"] = device_dtype.device
+    if device_dtype.dtype is not None:
+        kwargs["dtype"] = dtype_from_string[device_dtype.dtype]
+    return tensor.to(**kwargs)
 
 
 zero_tensor = torch.zeros(())
