@@ -8,6 +8,7 @@ use pyo3::{
     pyclass::CompareOp,
     types::{IntoPyDict, PyTuple},
 };
+use rustc_hash::FxHashMap as HashMap;
 
 use crate::{
     tensor_util::{Shape, TorchDeviceDtype},
@@ -35,6 +36,7 @@ pub struct PyUtils {
     pub builtins: Py<PyModule>,
     pub gc: Py<PyModule>,
     pub optimizing_symbolic_size_warning: PyObject,
+    pub maybe_dtype_to_maybe_string: PyObject,
 }
 
 /// misc python utilities
@@ -84,6 +86,7 @@ pub static PY_UTILS: GILLazyPy<PyUtils> = GILLazyPy::new_py(|py| {
         builtins: builtins.into(),
         gc: PyModule::import(py, "gc").unwrap().into(),
         optimizing_symbolic_size_warning: get("OptimizingSymbolicSizeWarning"),
+        maybe_dtype_to_maybe_string: get("maybe_dtype_to_maybe_string"),
     }
 });
 
