@@ -12,7 +12,10 @@ use macro_rules_attribute::apply;
 use pyo3::{
     exceptions,
     exceptions::PyValueError,
-    ffi, pyclass, pymethods,
+    ffi,
+    prelude::*,
+    pyclass::{self, CompareOp},
+    pymethods,
     types::{IntoPyDict, PySlice, PyTuple},
     AsPyPointer, FromPyObject, IntoPy, PyAny, PyErr, PyObject, PyResult, Python, ToPyObject,
 };
@@ -25,7 +28,7 @@ use super::py_types::{ExtraPySelfOps, PyShape};
 use crate::{
     atr, filter_by_variant,
     lru_cache::TensorCacheRrfs,
-    py_types::{Tensor, PY_UTILS},
+    py_types::{use_rust_comp, Tensor, PY_UTILS},
     pycall, python_error_exception, sv,
     tensor_db::{get_tensor_prefix, save_tensor},
     util::HashBytes,
@@ -162,6 +165,9 @@ pub enum MiscInputError {
     #[error("Shapes aren't broadcastable, {shapes:?} ({e_name})")]
     NotBroadcastable { shapes: Vec<Shape> },
 
+    #[error("GeneralFunction Index index dtype isnt i64 ({e_name})")]
+    IndexDtypeNotI64 {},
+
     #[error("item={item}, count={count} (needed -{count} <= {item} < {count}) ({e_name})")]
     ItemOutOfBounds { item: i64, count: usize },
 
@@ -171,6 +177,12 @@ pub enum MiscInputError {
         b: Option<TorchDtype>,
     },
 
+    #[error("Cast node got incompatible input, required compatibility {required:?}  actual {actual:?} ({e_name})")]
+    CastIncompatibleDeviceDtype {
+        required: TorchDeviceDtypeOp,
+        actual: TorchDeviceDtypeOp,
+    },
+
     #[error("Children multiple dtypes {a:?} {b:?} ({e_name})")]
     ChildrenMultipleDevices {
         a: Option<TorchDevice>,
@@ -932,13 +944,27 @@ pub fn string_escape(s: &String) -> String {
 /// use string so you can compare fast when constructing and have full range (weird types like uint or bfloat16)
 /// as opposed to PyObject torch.dtype or enum (could switch to enum)
 #[pyclass]
-#[derive(Debug, Clone, PartialEq, Eq, Default)]
+#[derive(Copy, Clone, PartialEq, Eq, Default, PartialOrd, Ord)]
 pub struct TorchDeviceDtypeOp {
     #[pyo3(get)]
     pub device: Option<TorchDevice>,
     #[pyo3(get)]
     pub dtype: Option<TorchDtype>,
 }
+impl std::fmt::Debug for TorchDeviceDtypeOp {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(
+            f,
+            "{{device:{},dtype:{}}}",
+            self.device
+                .map(|x| String::from(&x))
+                .unwrap_or("None".to_owned()),
+            self.dtype
+                .map(|x| String::from(&x))
+                .unwrap_or("None".to_owned())
+        )
+    }
+}
 
 #[pymethods]
 impl TorchDeviceDtypeOp {
@@ -949,7 +975,6 @@ impl TorchDeviceDtypeOp {
     }
 
     #[new]
-    #[args(device = "None", dtype = "None")]
     fn new(device: Option<TorchDevice>, dtype: Option<TorchDtype>) -> Self {
         Self { device, dtype }
     }
@@ -967,6 +992,10 @@ impl TorchDeviceDtypeOp {
                 .unwrap_or("None".to_owned())
         )
     }
+
+    fn __richcmp__(&self, object: &Self, comp_op: CompareOp) -> bool {
+        use_rust_comp(&self, &object, comp_op)
+    }
 }
 macro_rules! lit_or {
     ($a:ident, $b:literal) => {
@@ -977,7 +1006,7 @@ macro_rules! lit_or {
     };
 }
 macro_rules! string_enum{
-    {$name:ident, ($($var:ident $(=$lit:literal)?),*)}=>{
+    {$name:ident,$converter:expr, ($($var:ident $(=$lit:literal)?),*)}=>{
         #[derive(Copy, Clone, Debug, Hash, PartialEq, Eq, PartialOrd, Ord)]
         #[allow(non_camel_case_types)]
         pub enum $name{
@@ -1013,7 +1042,13 @@ macro_rules! string_enum{
         }
         impl<'source> FromPyObject<'source> for $name {
             fn extract(circuit_obj: &'source PyAny) -> PyResult<Self> {
-                let inp_str:String = circuit_obj.extract()?;
+                let converter:Option<&dyn Fn(&pyo3::PyAny)->Option<String>> = $converter;
+                let inp_str:String  = if let Some(converter) = converter{
+                    circuit_obj.extract().or_else(|_e|{let thing:Option<String> = converter(circuit_obj); thing.ok_or_else(||anyhow!("dtype not string or pytorch dtype object"))}
+                )
+                }else{
+                    circuit_obj.extract::<String>().context("string enum conversion failed")
+                }?;
                 let resulty:Result<$name> = inp_str.try_into();
                 resulty.map_err(|e|e.into())
             }
@@ -1022,6 +1057,7 @@ macro_rules! string_enum{
 }
 string_enum! {
     TorchDtype,
+    Some(&|obj|pycall!(PY_UTILS.maybe_dtype_to_maybe_string,(obj,))),
     (
         float16,
         bfloat16,
@@ -1038,6 +1074,7 @@ string_enum! {
 
 string_enum! {
     TorchDevice,
+    None,
     (
         Cpu="cpu",
         Cuda="cuda",
@@ -1061,13 +1098,23 @@ string_enum! {
 }
 
 #[pyclass]
-#[derive(Debug, Clone, PartialEq, Eq)]
+#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
 pub struct TorchDeviceDtype {
     #[pyo3(get)]
     pub device: TorchDevice,
     #[pyo3(get)]
     pub dtype: TorchDtype,
 }
+impl std::fmt::Debug for TorchDeviceDtype {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(
+            f,
+            "{{device:{},dtype:{}}}",
+            String::from(&self.device),
+            String::from(&self.dtype)
+        )
+    }
+}
 
 #[pymethods]
 impl TorchDeviceDtype {
@@ -1103,6 +1150,9 @@ impl TorchDeviceDtype {
             String::from(&self.dtype),
         )
     }
+    fn __richcmp__(&self, object: &Self, comp_op: CompareOp) -> bool {
+        use_rust_comp(&self, &object, comp_op)
+    }
 }
 
 impl TorchDeviceDtype {
@@ -1206,10 +1256,32 @@ impl TorchDeviceDtypeOp {
         }
         result
     }
+    pub fn override_other(&self, other: TorchDeviceDtypeOp) -> TorchDeviceDtypeOp {
+        let mut result = other.clone();
+        if let Some(x) = &self.device {
+            result.device = Some(*x);
+        }
+        if let Some(x) = &self.dtype {
+            result.dtype = Some(*x);
+        }
+        result
+    }
     pub fn cast_tensor(&self, tensor: Tensor) -> Tensor {
         let concrete = self.overwrite(&TorchDeviceDtype::from_tensor(&tensor));
         concrete.cast_tensor(tensor)
     }
+    pub fn hash(&self, hasher: &mut blake3::Hasher) {
+        if let Some(dev) = self.device {
+            hasher.update(&[1, dev as u8]);
+        } else {
+            hasher.update(&[0, 0]);
+        }
+        if let Some(typ) = self.dtype {
+            hasher.update(&[1, typ as u8]);
+        } else {
+            hasher.update(&[0, 0]);
+        }
+    }
 }
 
 impl Default for TorchDeviceDtype {
