@@ -560,15 +560,15 @@ where
 
 /// UNTESTED!
 pub fn fuse_maps<K, V, NewV, S>(
-    mut maps: Vec<std::collections::HashMap<K, V, S>>,
+    mut maps: Vec<::std::collections::HashMap<K, V, S>>,
     mut f: impl FnMut(Box<dyn Iterator<Item = V> + '_>) -> NewV,
-) -> std::collections::HashMap<K, NewV, S>
+) -> ::std::collections::HashMap<K, NewV, S>
 where
     S: Default + BuildHasher,
     K: Clone + Hash + Debug + Eq,
 {
     match split_first_take(&mut maps) {
-        None => std::collections::HashMap::default(),
+        None => ::std::collections::HashMap::default(),
         Some((first, rest)) => {
             let mut rest: Vec<_> = rest.collect();
             let first_keys: HashSet<_> = first.keys().cloned().collect();
